from __future__ import print_function
from distutils.version import LooseVersion
import ast
import os
import re
import sys

DEV_RELEASE = os.environ.get("DEV_RELEASE", None) == "1"
source_dir = sys.argv[1]

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('%s/__init__.py' % source_dir, 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

# Strip .devN
version_tuple = LooseVersion(version).version[0:3]
if not DEV_RELEASE:
    # Strip .devN
    version_tuple = LooseVersion(version).version[0:3]
    print(".".join(map(str, version_tuple)))
else:
    print(version)



import os
import json
import time

from six.moves.urllib.parse import quote


def test_standard_requests():
    """ Tests app controller methods. These tests should be
    compartmentalized. Also these methods should be made to not retest
    the behavior of the associated Manager class. """
    from .test_utils import test_pulsar_app

    with test_pulsar_app(test_conf={"extra_environ": {"REMOTE_ADDR": "127.101.101.98"}}) as app:
        staging_directory = app.app.staging_directory
        setup_response = app.post("/jobs?job_id=12345")
        setup_config = json.loads(setup_response.body.decode("utf-8"))
        assert setup_config["working_directory"].startswith(staging_directory)
        outputs_directory = setup_config["outputs_directory"]
        assert outputs_directory.startswith(staging_directory)
        assert setup_config["path_separator"] == os.sep
        job_id = setup_config["job_id"]

        def test_upload(upload_type):
            url = "/jobs/%s/files?name=input1&type=%s" % (job_id, upload_type)
            upload_input_response = app.post(url, "Test Contents")
            upload_input_config = json.loads(upload_input_response.body.decode("utf-8"))
            staged_input_path = upload_input_config["path"]
            staged_input = open(staged_input_path, "r")
            try:
                assert staged_input.read() == "Test Contents"
            finally:
                staged_input.close()
        test_upload("input")
        test_upload("tool")

        test_output = open(os.path.join(outputs_directory, "test_output"), "w")
        try:
            test_output.write("Hello World!")
        finally:
            test_output.close()
        download_response = app.get("/jobs/%s/files?name=test_output&type=output" % job_id)
        assert download_response.body.decode("utf-8") == "Hello World!"

        try:
            app.get("/jobs/%s/files?name=test_output2&type=output" % job_id)
            assert False  # Should throw exception
        except Exception:
            pass

        command_line = quote("""python -c "import sys; sys.stdout.write('test_out')" """)
        launch_response = app.post("/jobs/%s/submit?command_line=%s" % (job_id, command_line))
        assert launch_response.body.decode("utf-8") == 'OK'

        # Hack: Call twice to ensure postprocessing occurs and has time to
        # complete. Monitor thread should get this.
        time.sleep(.3)
        check_response = app.get("/jobs/%s/status" % job_id)
        time.sleep(.3)
        check_response = app.get("/jobs/%s/status" % job_id)
        check_config = json.loads(check_response.body.decode("utf-8"))
        assert check_config['returncode'] == 0
        assert check_config['stdout'] == "test_out"
        assert check_config['stderr'] == ""

        kill_response = app.put("/jobs/%s/cancel" % job_id)
        assert kill_response.body.decode("utf-8") == 'OK'

        clean_response = app.delete("/jobs/%s" % job_id)
        assert clean_response.body.decode("utf-8") == 'OK'
        assert os.listdir(staging_directory) == []



"""Tests for ``pulsar.client.amqp_exchange``."""
import time
import threading

from pulsar.client import amqp_exchange

from .test_utils import (
    skip_unless_module,
    timed,
)

TEST_CONNECTION = "memory://test_amqp"


@skip_unless_module("kombu")
@timed(15)
def test_amqp():
    """Test the client PulsarExchange abstraction with an in-memory connection."""
    manager1_exchange = amqp_exchange.PulsarExchange(TEST_CONNECTION, "manager_test")
    manager3_exchange = amqp_exchange.PulsarExchange(TEST_CONNECTION, "manager3_test")
    manager2_exchange = amqp_exchange.PulsarExchange(TEST_CONNECTION, "manager2_test")
    thread1 = TestThread("manager_test", manager1_exchange)
    thread2 = TestThread("manager2_test", manager2_exchange)
    thread3 = TestThread("manager3_test", manager3_exchange)
    thread1.start()
    thread2.start()
    thread3.start()
    time.sleep(0.5)
    manager1_exchange.publish("manager_test", u"cow1")
    manager2_exchange.publish("manager2_test", u"cow2")
    manager3_exchange.publish("manager3_test", u"cow3")
    time.sleep(0.1)
    thread1.wait_for_message(u"cow1")
    thread2.wait_for_message(u"cow2")
    thread3.wait_for_message(u"cow3")


class TestThread(threading.Thread):

    def __init__(self, queue_name, exchange):
        super(TestThread, self).__init__(target=self.run)
        self.queue_name = queue_name
        self.daemon = True
        self.exchange = exchange
        self.message = None

    def __nonzero__(self):
        return self.message is None

    __bool__ = __nonzero__  # Both needed Py2 v 3

    def run(self):
        def callback(body, message):
            self.message = body
            message.ack()

        self.exchange.consume(self.queue_name, callback=callback, check=self)

    def wait_for_message(self, expected_message):
        while self:
            time.sleep(.05)
        if self.message != expected_message:
            msg = "Expected [%s], got [%s]." % (expected_message, self.message)
            raise AssertionError(msg)

        self.join(2)


__all__ = ["test_amqp"]



from os import environ

from pulsar.client.manager import ClientManager


def test_environment_variables_config():
    environ['PULSAR_CACHE_TRANSFERS'] = '1'
    client_manager = ClientManager()
    assert __produces_caching_client(client_manager)

    environ['PULSAR_CACHE_TRANSFERS'] = '0'
    client_manager = ClientManager()
    assert not __produces_caching_client(client_manager)

    environ['PULSAR_CACHE_TRANSFERS'] = '1'
    environ['PULSAR_CACHE_THREADS'] = '4'
    client_manager = ClientManager()
    client_manager.client_cacher.num_transfer_threads == 4


def test_kwds_config():
    client_manager = ClientManager(cache=True, transfer_threads=3)
    assert __produces_caching_client(client_manager)
    assert client_manager.client_cacher.num_transfer_threads == 3

    client_manager = ClientManager(cache=False)
    assert not __produces_caching_client(client_manager)


def __produces_caching_client(client_manager):
    return client_manager.client_class.__name__.find('Caching') > 0



import os.path

from pulsar.client import PathMapper
from pulsar.client.action_mapper import path_type
from .test_utils import TempDirectoryTestCase

from galaxy.util.bunch import Bunch


class PathMapperTestCase(TempDirectoryTestCase):

    def test_input(self):
        local_path = os.path.join(os.path.dirname(self.temp_directory), "dataset1.dat")
        path_mapper = self._path_mapper(local_path, path_type.INPUT)
        new_path = path_mapper.remote_input_path_rewrite(local_path)
        assert new_path == "/scratch/staging/1/inputs/dataset1.dat"

    def test_output(self):
        local_path = os.path.join(os.path.dirname(self.temp_directory), "dataset1.dat")
        path_mapper = self._path_mapper(local_path, path_type.OUTPUT)
        new_path = path_mapper.remote_output_path_rewrite(local_path)
        assert new_path == "/scratch/staging/1/outputs/dataset1.dat"

    def test_output_workdir(self):
        local_path = os.path.join(self.temp_directory, "dataset1.dat")
        path_mapper = self._path_mapper(local_path, path_type.OUTPUT_WORKDIR)
        new_path = path_mapper.remote_output_path_rewrite(local_path)
        assert new_path == "/scratch/staging/1/working/dataset1.dat"

    def test_input_with_no_staging(self):
        local_path = os.path.join(os.path.dirname(self.temp_directory), "dataset1.dat")
        path_mapper = self._path_mapper(local_path, path_type.INPUT, staging_needed=False)
        new_path = path_mapper.remote_input_path_rewrite(local_path)
        assert new_path is None

    def test_output_with_no_staging(self):
        local_path = os.path.join(os.path.dirname(self.temp_directory), "dataset1.dat")
        path_mapper = self._path_mapper(local_path, path_type.OUTPUT, staging_needed=False)
        new_path = path_mapper.remote_output_path_rewrite(local_path)
        assert new_path is None

    def test_version_path(self):
        local_path = os.path.join(os.path.dirname(self.temp_directory), "GALAXY_VERSION_234")
        path_mapper = self._path_mapper(local_path, path_type.OUTPUT)
        new_path = path_mapper.remote_version_path_rewrite(local_path)
        assert new_path == "/scratch/staging/1/outputs/COMMAND_VERSION"

    def _path_mapper(self, expected_path, expected_type, staging_needed=True):
        action_mapper = TestActionMapper(expected_path, expected_type, staging_needed)
        path_mapper = PathMapper(
            client=None,
            remote_job_config=self.__test_remote_config(),
            local_working_directory=self.temp_directory,
            action_mapper=action_mapper,
        )
        return path_mapper

    def __test_remote_config(self):
        return dict(
            inputs_directory="/scratch/staging/1/inputs",
            outputs_directory="/scratch/staging/1/outputs",
            configs_directory="/scratch/staging/1/configs",
            working_directory="/scratch/staging/1/working",
            unstructured_files_directory="/scratch/staging/1/unstructured",
            system_properties=dict(separator="/"),
        )


class TestActionMapper(object):

    def __init__(self, expected_path, expected_type, staging_needed):
        self.expected_path = expected_path
        self.expected_type = expected_type
        self._action = Bunch(staging_needed=staging_needed)
        if not staging_needed:
            self._action.path_rewrite = lambda path: None

    def action(self, source, type):
        assert self.expected_path == source["path"]
        assert self.expected_type == type
        return self._action



from collections import deque
import os

from .test_utils import TempDirectoryTestCase
from pulsar.client.test.test_common import write_config
from pulsar.client import submit_job, ClientJobDescription
from pulsar.client import ClientOutputs
from galaxy.tool_util.deps.dependencies import DependenciesDescription
from galaxy.tool_util.deps.requirements import ToolRequirement

TEST_REQUIREMENT_1 = ToolRequirement("test1", "1.0")
TEST_REQUIREMENT_2 = ToolRequirement("test2", "1.0")
TEST_ENV_1 = dict(name="x", value="y")


class TestStager(TempDirectoryTestCase):

    def setUp(self):
        super(TestStager, self).setUp()
        from .test_utils import get_test_tool
        self.tool = get_test_tool()
        self.client = MockClient(self.temp_directory, self.tool)
        inputs = self.__setup_inputs()
        self.client_job_description = ClientJobDescription(
            tool=self.tool,
            command_line="run_test.exe",
            config_files=[],
            input_files=inputs,
            client_outputs=ClientOutputs("/galaxy/database/working_directory/1", []),
            working_directory="/galaxy/database/working_directory/1",
            dependencies_description=DependenciesDescription(requirements=[TEST_REQUIREMENT_1, TEST_REQUIREMENT_2]),
            env=[TEST_ENV_1],
            rewrite_paths=False,
        )
        self.job_config = dict(
            configs_directory="/pulsar/staging/1/configs",
            working_directory="/pulsar/staging/1/working",
            outputs_directory="/pulsar/staging/1/outputs",
            system_properties=dict(
                separator="\\",
            ),
        )

    def __setup_inputs(self):
        files_directory = os.path.join(self.temp_directory, "files")
        os.makedirs(files_directory)
        self.input1 = os.path.join(files_directory, "dataset_1.dat")
        self.input1_files_path = os.path.join(files_directory, "dataset_1_files")
        os.makedirs(self.input1_files_path)
        open(self.input1, "wb").write(b"012345")
        self.input2 = os.path.join(files_directory, "dataset_2.dat")
        open(self.input2, "wb").write(b"6789")
        return [self.input1, self.input2]

    def test_tool_file_rewrite(self):
        self.client_job_description.rewrite_paths = True
        tool_dir = os.path.abspath(self.tool.tool_dir)
        command_line = "python %s/tool1_wrapper.py" % tool_dir
        self.client_job_description.command_line = command_line
        rewritten_command_line = "python /pulsar/staging/1/tools/tool1_wrapper.py"
        self.client.expect_put_paths(["/pulsar/staging/1/tools/tool1_wrapper.py"])
        self.client.expect_command_line(rewritten_command_line)
        self._submit()
        uploaded_file1 = self.client.put_files[0]
        assert uploaded_file1[1] == "tool"
        self.assertEqual(uploaded_file1[0], "%s/tool1_wrapper.py" % tool_dir)

    def test_input_extra_rewrite(self):
        self.client_job_description.rewrite_paths = True
        extra_file = os.path.join(self.input1_files_path, "moo", "cow.txt")
        os.makedirs(os.path.dirname(extra_file))
        open(extra_file, "w").write("Hello World!")
        command_line = "test.exe %s" % extra_file
        self.client_job_description.command_line = command_line
        self.client.expect_command_line("test.exe /pulsar/staging/1/inputs/dataset_1_files/moo/cow.txt")
        self.client.expect_put_paths(["/pulsar/staging/1/inputs/dataset_1_files/moo/cow.txt"])
        self._submit()
        uploaded_file1 = self.client.put_files[0]
        assert uploaded_file1[1] == "input"
        assert uploaded_file1[0] == extra_file

    def test_unstructured_rewrite(self):
        self.client_job_description.rewrite_paths = True
        self.client.set_action_map_config(dict(paths=[
            dict(path=self.temp_directory, path_types="*any*")
        ]))
        local_unstructured_file = os.path.join(self.temp_directory, "A_RANDOM_FILE")
        open(local_unstructured_file, "wb").write(b"Hello World!")
        command_line = "foo.exe %s" % local_unstructured_file
        self.client_job_description.command_line = command_line
        self.client.expect_put_paths(["/pulsar/staging/1/other/A_RANDOM_FILE"])
        self.client.expect_command_line("foo.exe /pulsar/staging/1/other/A_RANDOM_FILE")
        self._submit()
        uploaded_file1 = self.client.put_files[0]
        assert uploaded_file1[1] == "unstructured"
        self.assertEqual(uploaded_file1[0], local_unstructured_file)

    def test_file_actions_by_dict(self):
        self.client_job_description.rewrite_paths = True
        self.client.set_action_map_config(dict(paths=[
            dict(path=self.temp_directory, path_types="*any*"),
        ]), by_path=False)
        local_unstructured_file = os.path.join(self.temp_directory, "A_RANDOM_FILE")
        open(local_unstructured_file, "wb").write(b"Hello World!")
        command_line = "foo.exe %s" % local_unstructured_file
        self.client_job_description.command_line = command_line
        self.client.expect_put_paths(["/pulsar/staging/1/other/A_RANDOM_FILE"])
        self.client.expect_command_line("foo.exe /pulsar/staging/1/other/A_RANDOM_FILE")
        self._submit()
        uploaded_file1 = self.client.put_files[0]
        assert uploaded_file1[1] == "unstructured"
        self.assertEqual(uploaded_file1[0], local_unstructured_file)

    def test_submit_no_rewrite(self):
        # Expect no rewrite of paths
        command_line_template = "run_test.exe --input1=%s --input2=%s"
        command_line = command_line_template % (self.input1, self.input2)
        self.client_job_description.command_line = command_line
        self.client.expect_command_line(command_line)
        self._submit()
        self._assert_inputs_uploaded()

    def test_submit_rewrite(self):
        self.client_job_description.rewrite_paths = True
        command_line_template = "run_test.exe --input1=%s --input2=%s"
        self.client_job_description.command_line = command_line_template % (self.input1, self.input2)
        rewritten_paths = (
            '/pulsar/staging/1/inputs/dataset_1.dat',
            '/pulsar/staging/1/inputs/dataset_2.dat',
        )
        self.client.expect_command_line(command_line_template % rewritten_paths)
        self._submit()
        self._assert_inputs_uploaded()

    def _assert_inputs_uploaded(self):
        # Expect both files staged
        uploaded_file1 = self.client.put_files[0]
        assert uploaded_file1[1] == "input"
        assert uploaded_file1[0] == self.input1
        uploaded_file2 = self.client.put_files[1]
        assert uploaded_file2[1] == "input"
        assert uploaded_file2[0] == self.input2

    def _submit(self):
        return submit_job(self.client, self.client_job_description, self.job_config)


class MockClient(object):

    def __init__(self, temp_directory, tool):
        self.temp_directory = temp_directory
        self.job_directory = None
        self.default_file_action = "transfer"
        self.action_config_path = None
        self.files_endpoint = None
        self.expected_tool = tool
        self.job_id = "1234"
        self.expected_command_line = None
        self.expect_put_paths([
            '/pulsar/staging/1/inputs/dataset_1.dat',
            '/pulsar/staging/1/inputs/dataset_2.dat',
        ])
        self.put_files = []

    def set_action_map_config(self, config, by_path=True):
        if by_path:
            self.action_config_path = write_config(self, config, name="actions.yaml")
        else:
            self.file_actions = config

    def expect_put_paths(self, paths):
        self.put_paths = deque(paths)

    def setup(self, tool_id, tool_version, use_metadata=False):
        assert tool_id == self.expected_tool.id
        assert tool_version == self.expected_tool.version
        return {}

    def launch(self, command_line, dependencies_description, job_config={}, remote_staging={}, env=[]):
        # TODO: test env
        if self.expected_command_line is not None:
            message = "Excepected command line %s, got %s" % (self.expected_command_line, command_line)
            assert self.expected_command_line == command_line, message
        assert dependencies_description.requirements == [TEST_REQUIREMENT_1, TEST_REQUIREMENT_2]
        assert env == [TEST_ENV_1]

    def expect_command_line(self, expected_command_line):
        self.expected_command_line = expected_command_line

    def put_file(self, path, type, name, contents, action_type='transfer'):
        self.put_files.append((path, type, name, contents))
        return {"path": self.put_paths.popleft()}



from os import makedirs
from os.path import join, dirname, exists
from string import Template
from galaxy.util.bunch import Bunch
from galaxy.objectstore import build_object_store_from_config

from .test_utils import TempDirectoryTestCase
from .test_utils import skip


class MockDataset(object):

    def __init__(self, id):
        self.id = id
        self.object_store_id = None


class PulsarObjectStoreTest(TempDirectoryTestCase):

    def __write(self, contents, name):
        path = join(self.temp_directory, name)
        directory = dirname(path)
        if not exists(directory):
            makedirs(directory)
        open(path, "wb").write(contents)
        return path

    @skip
    def test_pulsar_objectstore(self):
        # Define real object store used by Pulsar server.
        object_store_config_file = join(self.temp_directory, "object_store_conf.xml")
        with open(object_store_config_file, "w") as configf:
            config_template = Template("""<?xml version="1.0"?>
<object_store type="disk">
    <files_dir path="${temp_directory}"/>
    <extra_dir type="temp" path="${temp_directory}"/>
    <extra_dir type="job_work" path="${temp_directory}"/>
</object_store>
""")
            config_contents = config_template.safe_substitute(temp_directory=self.temp_directory)
            configf.write(config_contents)

        app_conf = dict(
            object_store_config_file=object_store_config_file,
            private_token="12345",
        )
        from .test_utils import test_pulsar_server
        with test_pulsar_server(app_conf=app_conf) as server:
            url = server.application_url
            # Define a proxy Pulsar object store.
            proxy_object_store_config_file = join(self.temp_directory, "proxy_object_store_conf.xml")
            with open(proxy_object_store_config_file, "w") as configf:
                config_template = Template("""<?xml version="1.0"?>
<object_store type="pulsar" url="$url" private_token="12345" transport="urllib">
  <!-- private_token is optional - see Pulsar documentation for more information. -->
  <!-- transport is optional, set to curl to use libcurl instead of urllib for communication with Pulsar. -->
</object_store>
""")
                contents = config_template.safe_substitute(url=url)
                configf.write(contents)

            config = Bunch(object_store_config_file=proxy_object_store_config_file)
            object_store = build_object_store_from_config(config=config)

            # Test no dataset with id 1 exists.
            absent_dataset = MockDataset(1)
            assert not object_store.exists(absent_dataset)

            # Write empty dataset 2 in second backend, ensure it is empty and
            # exists.
            empty_dataset = MockDataset(2)
            self.__write(b"", "000/dataset_2.dat")
            assert object_store.exists(empty_dataset)
            assert object_store.empty(empty_dataset)

            # Write non-empty dataset in backend 1, test it is not emtpy & exists.
            hello_world_dataset = MockDataset(3)
            self.__write(b"Hello World!", "000/dataset_3.dat")
            assert object_store.exists(hello_world_dataset)
            assert not object_store.empty(hello_world_dataset)

            # Test get_data
            data = object_store.get_data(hello_world_dataset)
            assert data == "Hello World!"

            data = object_store.get_data(hello_world_dataset, start=1, count=6)
            assert data == "ello W"

            # Test Size

            # Test absent and empty datasets yield size of 0.
            assert object_store.size(absent_dataset) == 0
            assert object_store.size(empty_dataset) == 0
            # Elsewise
            assert object_store.size(hello_world_dataset) > 0  # Should this always be the number of bytes?

            # Test percent used (to some degree)
            percent_store_used = object_store.get_store_usage_percent()
            assert percent_store_used > 0.0
            assert percent_store_used < 100.0

            # Test update_from_file test
            output_dataset = MockDataset(4)
            output_real_path = join(self.temp_directory, "000", "dataset_4.dat")
            assert not exists(output_real_path)
            output_working_path = self.__write(b"NEW CONTENTS", "job_working_directory1/example_output")
            object_store.update_from_file(output_dataset, file_name=output_working_path, create=True)
            assert exists(output_real_path)

            # Test delete
            to_delete_dataset = MockDataset(5)
            to_delete_real_path = self.__write(b"content to be deleted!", "000/dataset_5.dat")
            assert object_store.exists(to_delete_dataset)
            assert object_store.delete(to_delete_dataset)
            assert not object_store.exists(to_delete_dataset)
            assert not exists(to_delete_real_path)

            # Test json content.
            complex_contents_dataset = MockDataset(6)
            complex_content = b'{"a":6}'
            self.__write(complex_content, "000/dataset_6.dat")
            assert object_store.exists(complex_contents_dataset)
            data = object_store.get_data(complex_contents_dataset) == complex_content



from galaxy.util.bunch import Bunch
from pulsar.client.action_mapper import (
    FileActionMapper,
)


def test_endpoint_validation():
    client = _min_client("remote_transfer")
    mapper = FileActionMapper(client)
    exception_found = False
    try:
        mapper.action({'path': '/opt/galaxy/tools/filters/catWrapper.py'}, 'input')
    except Exception as e:
        exception_found = True
        assert "files_endpoint" in str(e)
    assert exception_found


def test_ssh_key_validation():
    client = _min_client("remote_rsync_transfer")
    mapper = FileActionMapper(client)
    exception_found = False
    try:
        mapper.action({'path': '/opt/galaxy/tools/filters/catWrapper.py'}, 'input')
    except Exception as e:
        exception_found = True
        assert "ssh_key" in str(e)
    assert exception_found


def test_ssh_key_defaults():
    client = _client("remote_rsync_transfer")
    mapper = FileActionMapper(client)
    action = mapper.action({'path': '/opt/galaxy/tools/filters/catWrapper.py'}, 'input')
    action.to_dict()


def _min_client(default_action):
    """Minimal client, missing properties for certain actions."""
    mock_client = Bunch(
        default_file_action=default_action,
        action_config_path=None,
        files_endpoint=None,
        ssh_key=None,
    )
    return mock_client


def _client(default_action):
    mock_client = Bunch(
        default_file_action=default_action,
        action_config_path=None,
        files_endpoint="http://localhost",
        ssh_key="12345",
    )
    return mock_client



import os

from .test_utils import files_server
from pulsar.client.action_mapper import RemoteTransferAction


def test_write_to_file():
    with files_server() as (server, directory):
        from_path = os.path.join(directory, "remote_get")
        open(from_path, "wb").write(b"123456")

        to_path = os.path.join(directory, "local_get")
        url = server.application_url + "?path=%s" % from_path
        RemoteTransferAction({"path": to_path}, url=url).write_to_path(to_path)

        assert open(to_path, "rb").read() == b"123456"


def test_write_from_file():
    with files_server() as (server, directory):
        from_path = os.path.join(directory, "local_post")
        open(from_path, "wb").write(b"123456")

        to_path = os.path.join(directory, "remote_post")
        url = server.application_url + "?path=%s" % to_path
        RemoteTransferAction({"path": to_path}, url=url).write_from_path(from_path)

        posted_contents = open(to_path, "rb").read()
        assert posted_contents == b"123456", posted_contents



#!/usr/bin/env python
# Modify version...
import datetime
import os
import re
import subprocess
import sys


DEV_RELEASE = os.environ.get("DEV_RELEASE", None) == "1"
PROJECT_DIRECTORY = os.path.join(os.path.dirname(__file__), "..")


def main(argv):
    source_dir = argv[1]
    version = argv[2]
    mod_path = os.path.join(PROJECT_DIRECTORY, source_dir, "__init__.py")
    if not DEV_RELEASE:
        history_path = os.path.join(PROJECT_DIRECTORY, "HISTORY.rst")
        history = open(history_path, "r").read()
        today = datetime.datetime.today()
        today_str = today.strftime('%Y-%m-%d')
        history = re.sub(r"^[\d\.]+\.dev\d*",
                         "{} ({})".format(version, today_str),
                         history,
                         flags=re.MULTILINE)
        open(history_path, "w").write(history)

        mod = open(mod_path, "r").read()
        mod = re.sub(r"__version__ = '[\d\.]+\.dev\d*'",
                     "__version__ = '{}'".format(version),
                     mod,
                     flags=re.MULTILINE)
        mod = open(mod_path, "w").write(mod)
    for f in ("HISTORY.rst", os.path.join(source_dir, "__init__.py")):
        if not shell(["git", "diff", "--quiet", "--", f]):
            print("Expected changes to '{}' not found, refusing to commit".format(f))
            sys.exit(1)
    shell(["git", "commit", "-m", "Version %s" % version,
           "HISTORY.rst", "%s/__init__.py" % source_dir])
    shell(["git", "tag", version])


def shell(cmds, **kwds):
    p = subprocess.Popen(cmds, **kwds)
    return p.wait()


if __name__ == "__main__":
    main(sys.argv)



from __future__ import print_function
from os.path import join
from os import environ, makedirs, system
from six import next, itervalues
from six.moves import configparser
from .test_utils import (
    TempDirectoryTestCase,
    skip_unless_executable,
    skip_unless_module,
    skip_unless_any_module,
    skip_unless_environ,
    skip_without_drmaa,
)

from .test_utils import integration_test
from .test_utils import test_pulsar_app
from .test_utils import test_pulsar_server
from .test_utils import files_server

from galaxy.util.bunch import Bunch
from pulsar.client.test.check import run


class BaseIntegrationTest(TempDirectoryTestCase):

    def _run(self, app_conf={}, job_conf_props={}, **kwds):
        app_conf = app_conf.copy()
        job_conf_props = job_conf_props.copy()

        if "suppress_output" not in kwds:
            kwds["suppress_output"] = False

        self.__setup_job_properties(app_conf, job_conf_props)
        self.__setup_dependencies(app_conf)
        self._run_in_app(app_conf, **kwds)

    def _run_in_app(self, app_conf, **kwds):
        direct_interface = kwds.get("direct_interface", None)
        inject_files_endpoint = direct_interface or kwds.get("inject_files_endpoint", False)
        if inject_files_endpoint:
            # Client directory hasn't bee created yet, don't restrict where
            # test files written.
            # Can only run tests using files_server if not constructing a test
            # server for Pulsar - webtest doesn't seem to like having two test
            # servers alive at same time.
            with files_server("/") as test_files_server:
                files_endpoint = to_infrastructure_uri(test_files_server.application_url)
                if direct_interface:
                    self._run_direct(app_conf, files_endpoint=files_endpoint, **kwds)
                else:
                    self._run_in_test_server(app_conf, files_endpoint=files_endpoint, **kwds)
        else:
            self._run_in_test_server(app_conf, **kwds)

    def _run_in_test_server(self, app_conf, **kwds):
        with test_pulsar_server(app_conf=app_conf) as server:
            options = Bunch(url=server.application_url, **kwds)
            # TODO: sync Py 2 v 3 approach so following hack is unneeded.
            app = server.test_app
            if hasattr(app, "application"):
                app = app.application
            self._update_options_for_app(options, app, **kwds)
            run(options)

    def _run_direct(self, app_conf, **kwds):
        with test_pulsar_app({}, app_conf, {}) as app:
            options = Bunch(job_manager=next(itervalues(app.app.managers)), file_cache=app.app.file_cache, **kwds)
            self._update_options_for_app(options, app.app, **kwds)
            run(options)

    def _update_options_for_app(self, options, app, **kwds):
        if kwds.get("local_setup", False):
            staging_directory = app.staging_directory
            if kwds.get("k8s_enabled"):
                options["jobs_directory"] = "/pulsar_staging"
            else:
                options["jobs_directory"] = staging_directory

    def __setup_job_properties(self, app_conf, job_conf_props):
        if job_conf_props:
            job_conf = join(self.temp_directory, "job_managers.ini")
            config = configparser.ConfigParser()
            section_name = "manager:_default_"
            config.add_section(section_name)
            for key, value in job_conf_props.items():
                config.set(section_name, key, value)
            with open(job_conf, "w") as configf:
                config.write(configf)

            app_conf["job_managers_config"] = job_conf

    def __setup_dependencies(self, app_conf):
        dependencies_dir = join(self.temp_directory, "dependencies")
        dep1_directory = join(dependencies_dir, "dep1", "1.1")
        makedirs(dep1_directory)
        try:
            # Let external users read/execute this directory for run as user
            # test.
            system("chmod 755 %s" % self.temp_directory)
            system("chmod -R 755 %s" % dependencies_dir)
        except Exception as e:
            print(e)
        env_file = join(dep1_directory, "env.sh")
        with open(env_file, "w") as env:
            env.write("MOO=moo_override; export MOO")
        app_conf["tool_dependency_dir"] = dependencies_dir


class IntegrationTests(BaseIntegrationTest):
    default_kwargs = dict(direct_interface=False, test_requirement=True, test_unicode=True, test_env=True, test_rewrite_action=True)

    @integration_test
    def test_integration_no_requirement(self):
        self._run(private_token=None, **self.default_kwargs)

    @integration_test
    def test_integration_maximum_stream_size(self):
        app_conf = dict(maximum_stream_size=4)
        self._run(app_conf=app_conf, private_token=None, maximum_stream_size=4, **self.default_kwargs)

    @skip_without_drmaa
    @integration_test
    def test_integration_as_user(self):
        job_props = {'type': 'queued_external_drmaa', "production": "false"}
        self._run(job_conf_props=job_props, private_token=None, default_file_action="copy", user='u1', **self.default_kwargs)

    @integration_test
    def test_integration_local_setup(self):
        self._run(private_token=None, default_file_action="remote_copy", local_setup=True, **self.default_kwargs)

    @skip_unless_module("pycurl")
    @skip_unless_module("kombu")
    @integration_test
    def test_message_queue(self):
        self._run(
            app_conf=dict(message_queue_url="memory://test1"),
            private_token=None,
            default_file_action="remote_copy",
            local_setup=True,
            manager_url="memory://test1",
            **self.default_kwargs
        )

    @skip_unless_environ("PULSAR_TEST_KEY")
    @skip_unless_module("kombu")
    @integration_test
    def test_integration_scp(self):
        self._run(
            app_conf=dict(message_queue_url="memory://test2"),
            private_token=None,
            default_file_action="remote_scp_transfer",
            local_setup=True,
            manager_url="memory://test2",
            **self.default_kwargs
        )

    @skip_unless_environ("PULSAR_TEST_KEY")
    @skip_unless_module("kombu")
    @integration_test
    def test_integration_rsync(self):
        self._run(
            app_conf=dict(message_queue_url="memory://test3"),
            private_token=None,
            default_file_action="remote_rsync_transfer",
            local_setup=True,
            manager_url="memory://test3",
            **self.default_kwargs
        )

    @integration_test
    def test_integration_copy(self):
        self._run(private_token=None, default_file_action="copy", **self.default_kwargs)

    @integration_test
    def test_integration_no_transfer(self):
        self._run(private_token=None, default_file_action="none", **self.default_kwargs)

    @integration_test
    def test_integration_cached(self):
        self._run(private_token=None, cache=True, **self.default_kwargs)

    @integration_test
    def test_integration_default(self):
        self._run(private_token=None, **self.default_kwargs)

    @skip_unless_module("pycurl")
    @integration_test
    def test_integration_curl(self):
        self._run(private_token=None, transport="curl", **self.default_kwargs)

    @integration_test
    def test_integration_explicit_tool_directory_includes(self):
        self._run(private_token=None, explicit_tool_declarations=True, **self.default_kwargs)

    @integration_test
    def test_integration_token(self):
        self._run(app_conf={"private_token": "testtoken"}, private_token="testtoken", **self.default_kwargs)

    @integration_test
    def test_integration_errors(self):
        self._run(app_conf={"private_token": "testtoken"}, private_token="testtoken", test_errors=True, **self.default_kwargs)

    @skip_without_drmaa
    @integration_test
    def test_integration_drmaa(self):
        self._run(app_conf={}, job_conf_props={'type': 'queued_drmaa'}, private_token=None, **self.default_kwargs)

    @skip_unless_executable("condor_submit")
    @integration_test
    def test_integration_condor(self):
        self._run(app_conf={}, job_conf_props={'type': 'queued_condor'}, private_token=None, **self.default_kwargs)

    @skip_unless_executable("qsub")
    @integration_test
    def test_integration_cli_torque(self):
        self._run(app_conf={}, job_conf_props={'type': 'queued_cli', 'job_plugin': 'Torque'}, private_token=None, **self.default_kwargs)

    @skip_unless_executable("sbatch")
    @integration_test
    def test_integration_cli_slurm(self):
        self._run(app_conf={}, job_conf_props={'type': 'queued_cli', 'job_plugin': 'Slurm'}, private_token=None, **self.default_kwargs)


class ExternalQueueIntegrationTests(IntegrationTests):
    default_kwargs = dict(direct_interface=False, test_requirement=False, test_unicode=True, test_env=True, test_rewrite_action=True)

    @integration_test
    @skip_unless_environ("PULSAR_RABBIT_MQ_CONNECTION")
    def test_integration_external_rabbit(self):
        # e.g. amqp://guest:guest@localhost:5672//
        # TODO: nc docker.for.mac.localhost 5679
        message_queue_url = environ.get("PULSAR_RABBIT_MQ_CONNECTION")
        self._run(
            app_conf=dict(message_queue_url=message_queue_url),
            private_token=None,
            local_setup=True,
            default_file_action="remote_transfer",
            manager_url=message_queue_url,
            inject_files_endpoint=True,
            **self.default_kwargs
        )

    @integration_test
    @skip_unless_environ("PULSAR_RABBIT_MQ_CONNECTION")
    def test_integration_kubernetes(self):
        message_queue_url = environ.get("PULSAR_RABBIT_MQ_CONNECTION")
        remote_pulsar_app_config = {
            "staging_directory": "/pulsar_staging/",
            "message_queue_url": to_infrastructure_uri(message_queue_url),
            "manager": {
                "type": "coexecution",
            }
        }
        self._run(
            app_conf=dict(message_queue_url=message_queue_url),
            private_token=None,
            local_setup=True,
            default_file_action="remote_transfer",
            manager_url=message_queue_url,
            inject_files_endpoint=True,
            k8s_enabled=True,
            container="conda/miniconda2",
            remote_pulsar_app_config=remote_pulsar_app_config,
            **self.default_kwargs
        )


class DirectIntegrationTests(IntegrationTests):
    default_kwargs = dict(direct_interface=True, test_requirement=False)

    @skip_unless_any_module(["pycurl", "poster", "requests_toolbelt"])
    @integration_test
    def test_integration_remote_transfer(self):
        self._run(
            private_token=None,
            local_setup=True,
            default_file_action="remote_transfer",
            **self.default_kwargs
        )


def to_infrastructure_uri(uri):
    # remap MQ or file server URI hostnames for in-container versions, this is sloppy
    # should actually parse the URI and rebuild with correct host
    infrastructure_host = environ.get("PULSAR_TEST_INFRASTRUCTURE_HOST")
    infrastructure_uri = uri
    if infrastructure_host:
        if "0.0.0.0" in infrastructure_uri:
            infrastructure_uri = infrastructure_uri.replace("0.0.0.0", infrastructure_host)
        elif "localhost" in infrastructure_uri:
            infrastructure_uri = infrastructure_uri.replace("localhost", infrastructure_host)
        elif "127.0.0.1" in infrastructure_uri:
            infrastructure_uri = infrastructure_uri.replace("127.0.0.1", infrastructure_host)
    return infrastructure_uri



from contextlib import contextmanager
from os.path import exists, join
import time

from pulsar.managers.queued import QueueManager
from pulsar.managers.stateful import StatefulManagerProxy
from pulsar.tools.authorization import get_authorizer
from .test_utils import (
    temp_directory,
    TestDependencyManager
)
from galaxy.job_metrics import NULL_JOB_INSTRUMENTER
from galaxy.util.bunch import Bunch

TEST_JOB_ID = "4"
TEST_STAGED_FILE = "cow"
TEST_COMMAND_TOUCH_FILE = "ran"


def test_launched_job_recovery():
    """Tests persistence and recovery of launched managers jobs."""
    with _app() as app:
        staging_directory = app.staging_directory
        queue1 = StatefulManagerProxy(QueueManager('test', app, num_concurrent_jobs=0))
        job_id = queue1.setup_job(TEST_JOB_ID, 'tool1', '1.0.0')
        touch_file = join(staging_directory, TEST_COMMAND_TOUCH_FILE)
        queue1.preprocess_and_launch(job_id, {"command_line": 'touch %s' % touch_file})
        time.sleep(.4)
        assert not exists(touch_file)
        queue1.shutdown()
        _setup_manager_that_executes(app)
        assert exists(touch_file)


def test_preprocessing_job_recovery():
    """Tests persistence and recovery of preprocessing managers jobs (clean)."""
    with _app() as app:
        _setup_job_with_unexecuted_preprocessing_directive(app)
        staging_directory = app.staging_directory
        staged_file = join(staging_directory, TEST_JOB_ID, "inputs", TEST_STAGED_FILE)
        touch_file = join(staging_directory, TEST_COMMAND_TOUCH_FILE)

        # File shouldn't have been staged because we hacked stateful proxy manager to not
        # run preprocess.
        assert not exists(staged_file)

        _setup_manager_that_preprocesses(app)

        assert exists(staged_file)
        assert not exists(touch_file)

        _setup_manager_that_executes(app)
        assert exists(touch_file)


def test_preprocessing_job_recovery_dirty():
    """Tests persistence and recovery of preprocessing managers jobs (dirty)."""

    # Same test as above, but simulating existing files from a previous partial
    # preprocess.
    with _app() as app:
        _setup_job_with_unexecuted_preprocessing_directive(app)
        staging_directory = app.staging_directory
        staged_file = join(staging_directory, TEST_JOB_ID, "inputs", TEST_STAGED_FILE)
        touch_file = join(staging_directory, TEST_COMMAND_TOUCH_FILE)

        # File shouldn't have been staged because we hacked stateful proxy manager to not
        # run preprocess.
        assert not exists(staged_file)
        # write out partial contents, make sure preprocess writes over this with the correct
        # contents.
        open(staged_file, "wb").write(b"co")
        _setup_manager_that_preprocesses(app)

        assert exists(staged_file)
        assert open(staged_file, "rb").read() == b"cow file"
        assert not exists(touch_file)

        _setup_manager_that_executes(app)
        assert exists(touch_file)


def _setup_manager_that_preprocesses(app):
    # Setup a manager that will preprocess the job but won't execute it.

    # Now start a real stateful manager proxy and watch the file get staged.
    queue2 = StatefulManagerProxy(QueueManager('test', app, num_concurrent_jobs=0))
    try:
        queue2.recover_active_jobs()
        time.sleep(1)
    finally:
        try:
            queue2.shutdown()
        except Exception:
            pass


def _setup_job_with_unexecuted_preprocessing_directive(app):
    staging_directory = app.staging_directory
    queue1 = DoesntPreprocessStatefulManagerProxy(QueueManager('test', app, num_concurrent_jobs=0))
    job_id = queue1.setup_job(TEST_JOB_ID, 'tool1', '1.0.0')
    action = {"name": TEST_STAGED_FILE, "type": "input", "action": {"action_type": "message", "contents": "cow file"}}
    remote_staging = {"setup": [action]}
    touch_file = join(staging_directory, TEST_COMMAND_TOUCH_FILE)
    queue1.preprocess_and_launch(job_id, {"command_line": "touch '%s'" % touch_file, "remote_staging": remote_staging})
    queue1.shutdown()


def _setup_manager_that_executes(app):
    queue2 = StatefulManagerProxy(QueueManager('test', app, num_concurrent_jobs=1))
    try:
        queue2.recover_active_jobs()
        time.sleep(1)
    finally:
        try:
            queue2.shutdown()
        except Exception:
            pass


@contextmanager
def _app():
    with temp_directory() as staging_directory:
        app = Bunch(
            staging_directory=staging_directory,
            persistence_directory=staging_directory,
            authorizer=get_authorizer(None),
            dependency_manager=TestDependencyManager(),
            job_metrics=Bunch(default_job_instrumenter=NULL_JOB_INSTRUMENTER),
            object_store=None,
        )
        yield app


class DoesntPreprocessStatefulManagerProxy(StatefulManagerProxy):

    def _launch_prepreprocessing_thread(self, job_id, launch_config):
        pass



from .test_utils import TempDirectoryTestCase
from pulsar.managers.base import JobDirectory
import os

TEST_JOB_ID = "1234"


class JobDirectoryTestCase(TempDirectoryTestCase):

    def setUp(self):
        super(JobDirectoryTestCase, self).setUp()
        self.job_directory = JobDirectory(self.temp_directory, TEST_JOB_ID)

    def test_setup(self):
        expected_path = os.path.join(self.temp_directory, TEST_JOB_ID)
        assert not os.path.exists(expected_path)
        self.job_directory.setup()
        assert os.path.exists(expected_path)

    def test_metadata(self):
        self.prep()
        assert not self.job_directory.has_metadata("MooCow")
        self.job_directory.store_metadata("MooCow", True)
        assert self.job_directory.has_metadata("MooCow")

    def prep(self):
        self.job_directory.setup()



import pulsar.scripts.chown_working_directory
import pulsar.scripts.drmaa_kill
import pulsar.scripts.drmaa_launch
import pulsar.scripts.mesos_executor
import pulsar.scripts.mesos_framework
import pulsar.scripts.submit
import pulsar.client.test.check

MODULES = [
    pulsar.scripts.drmaa_kill,
    pulsar.scripts.drmaa_launch,
    pulsar.scripts.mesos_executor,
    pulsar.scripts.mesos_framework,
    pulsar.scripts.submit,
    pulsar.client.test.check,
]


def test_scripts_have_help():
    for module in MODULES:
        try:
            module.main(["-h"])
        except SystemExit as e:
            assert e.code == 0, module



import os
from tempfile import NamedTemporaryFile

from pulsar.client.transport.standard import UrllibTransport
from pulsar.client.transport.curl import PycurlTransport
from pulsar.client.transport.curl import post_file
from pulsar.client.transport.curl import get_file
from pulsar.client.transport import get_transport

from .test_utils import files_server
from .test_utils import skip_unless_module


def test_urllib_transports():
    _test_transport(UrllibTransport())


@skip_unless_module("pycurl")
def test_pycurl_transport():
    _test_transport(PycurlTransport())


def _test_transport(transport):
    with files_server(allow_multiple_downloads=True) as (server, directory):
        path = os.path.join(directory, "test_for_GET")
        open(path, "w").write(" Test123 ")

        server_url = server.application_url
        request_url = u"%s?path=%s" % (server_url, path)

        # Testing simple get
        response = transport.execute(request_url, data=None)
        assert response.find(b"Test123") >= 0

        # Testing writing to output file
        temp_file = NamedTemporaryFile(delete=True)
        output_path = temp_file.name
        temp_file.close()
        response = transport.execute(request_url, data=None, output_path=output_path)
        assert open(output_path, 'r').read().find("Test123") >= 0


@skip_unless_module("pycurl")
def test_curl_put_get():
    with files_server() as (server, directory):
        server_url = server.application_url
        path = os.path.join(directory, "test_for_GET")
        request_url = u"%s?path=%s" % (server_url, path)

        input = os.path.join(directory, "input")
        output = os.path.join(directory, "output")
        open(input, "w").write(u"helloworld")

        post_file(request_url, input)
        get_file(request_url, output)
        assert open(output, "r").read() == u"helloworld"


def test_curl_status_code():
    with files_server() as (server, directory):
        server_url = server.application_url
        path = os.path.join(directory, "test_for_GET")
        request_url = u"%s?path=%s" % (server_url, path)
        # Verify curl doesn't just silently swallow errors.
        exception_raised = False
        try:
            get_file(request_url, os.path.join(directory, "test"))
        except Exception:
            exception_raised = True
        assert exception_raised

        post_request_url = u"%s?path=%s" % (server_url, "/usr/bin/cow")
        exception_raised = False
        try:
            post_file(post_request_url, os.path.join(directory, "test"))
        except Exception:
            exception_raised = True
        assert exception_raised


@skip_unless_module("pycurl")
def test_curl_problems():
    with files_server() as (server, directory):
        server_url = server.application_url
        path = os.path.join(directory, "test_for_GET")
        request_url = u"%s?path=%s" % (server_url, path)
        exception_raised = False
        try:
            # Valid destination but the file to post doesn't exist.
            post_file(request_url, os.path.join(directory, "test"))
        except Exception:
            exception_raised = True
        assert exception_raised


def test_get_transport():
    assert type(get_transport(None, FakeOsModule("1"))) == PycurlTransport
    assert type(get_transport(None, FakeOsModule("TRUE"))) == PycurlTransport
    assert type(get_transport(None, FakeOsModule("0"))) == UrllibTransport
    assert type(get_transport('urllib', FakeOsModule("TRUE"))) == UrllibTransport
    assert type(get_transport('curl', FakeOsModule("TRUE"))) == PycurlTransport


class FakeOsModule(object):

    def __init__(self, env_val):
        self.env_val = env_val

    def getenv(self, key, default):
        return self.env_val



from pulsar.managers.util.retry import RetryActionExecutor


def test_retry_defaults():
    action_tracker = ActionTracker()
    assert RetryActionExecutor().execute(action_tracker.execute) == 42
    assert action_tracker.count == 1


def test_exception_passthrough():
    action_tracker = ActionTracker(fail_count=1)
    exception_raised = False
    try:
        RetryActionExecutor().execute(action_tracker.execute)
    except Exception:
        exception_raised = True
    assert action_tracker.count == 1
    assert exception_raised


def test_third_execution_fine():
    action_tracker = ActionTracker(fail_count=2)
    exception_raised = False
    try:
        RetryActionExecutor(max_retries=2, interval_start=.01, interval_step=.01).execute(action_tracker.execute)
    except Exception:
        exception_raised = True
    assert action_tracker.count == 3, action_tracker.count
    assert not exception_raised


class ActionTracker(object):

    def __init__(self, fail_count=0, fail_how=Exception):
        self.fail_count = fail_count
        self.fail_how = fail_how
        self.count = 0

    def execute(self):
        self.count += 1
        if self.fail_count >= self.count:
            raise self.fail_how()
        else:
            return 42



from .test_utils import get_test_toolbox, test_job_directory
from os.path import join
from os import makedirs


def test_load_simple_tool():
    toolbox = get_test_toolbox()
    tool1 = toolbox.get_tool("tool1")
    assert tool1.version == "0.1"


def test_command_validation():
    toolbox = get_test_toolbox()
    tool1 = toolbox.get_tool("tool1")
    with test_job_directory() as job_directory:
        wrapper_path = join(job_directory.tool_files_directory(), "tool1_wrapper.py")
        config_path = join(job_directory.configs_directory(), "config1")
        valid_command = "python %s --config %s" % \
            (wrapper_path, config_path)
        assert tool1.inputs_validator.validate_command(job_directory, valid_command)


def test_config_validation():
    toolbox = get_test_toolbox()
    tool1 = toolbox.get_tool("tool1")
    with test_job_directory() as job_directory:
        configs_dir = job_directory.configs_directory()
        config_path = join(configs_dir, "config1")
        makedirs(configs_dir)
        open(config_path, "w").write("\n   --option1='Hello' \n")
        assert tool1.inputs_validator.validate_config(job_directory, "config1", config_path)
